/* Copyright 2013 Richard Taylor

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.

  CREDITS:

  Much of the code here was adapted from Steve Harris's swh-plugins
  (version 0.4.15) found at http://plugin.org.uk/.  Thanks Steve for
  making your code available.

  The guts of the code, in util/biquad.h, is an implementation
  of the biquad filters described in Robert Bristow-Johnson's "Cookbook
  formulae for audio EQ biquad filter coefficients" found here:
  http://www.musicdsp.org/files/Audio-EQ-Cookbook.txt
*/

#include <stdlib.h>
#include <string.h>
#include <xmmintrin.h>

#define         _ISOC9X_SOURCE  1
#define         _ISOC99_SOURCE  1
#define         __USE_ISOC99    1
#define         __USE_ISOC9X    1

#include <math.h>

#include "ladspa.h"

#include "util/biquad.h"

#define ALLPASS_FC                  0
#define ALLPASS_BW                  1
#define ALLPASS_INPUT               2
#define ALLPASS_OUTPUT              3

static LADSPA_Descriptor *allPassDescriptor = NULL;

typedef struct {
	LADSPA_Data *gain;
	LADSPA_Data *fc;
	LADSPA_Data *bw;
	LADSPA_Data *input;
	LADSPA_Data *output;
	biquad *     filter;
	float        fs;
} AllPass;

const LADSPA_Descriptor *ladspa_descriptor(unsigned long index) {
	switch (index) {
	case 0:
		return allPassDescriptor;
	default:
		return NULL;
	}
}

static void activateAllPass(LADSPA_Handle instance) {
	AllPass *plugin_data = (AllPass *)instance;
	biquad *filter = plugin_data->filter;
	float fs = plugin_data->fs;
	biquad_init(filter);
	plugin_data->filter = filter;
	plugin_data->fs = fs;

}

static void cleanupAllPass(LADSPA_Handle instance) {
	AllPass *plugin_data = (AllPass *)instance;
	free(plugin_data->filter);
	free(instance);
}

static void connectPortAllPass(
 LADSPA_Handle instance,
 unsigned long port,
 LADSPA_Data *data) {
	AllPass *plugin;

	plugin = (AllPass *)instance;
	switch (port) {
	case ALLPASS_FC:
		plugin->fc = data;
		break;
	case ALLPASS_BW:
		plugin->bw = data;
		break;
	case ALLPASS_INPUT:
		plugin->input = data;
		break;
	case ALLPASS_OUTPUT:
		plugin->output = data;
		break;
	}
}

static LADSPA_Handle instantiateAllPass(
 const LADSPA_Descriptor *descriptor,
 unsigned long s_rate) {
	AllPass *plugin_data = (AllPass *)malloc(sizeof(AllPass));
	biquad *filter = NULL;
	float fs;

	fs = (float)s_rate;
	filter = malloc(sizeof(biquad));
	biquad_init(filter);

	plugin_data->filter = filter;
	plugin_data->fs = fs;

	return (LADSPA_Handle)plugin_data;
}

#undef buffer_write
#define buffer_write(b, v) (b = v)

static void runAllPass(LADSPA_Handle instance, unsigned long sample_count) {
	AllPass *plugin_data = (AllPass *)instance;

	/* Frequency (Hz) (float value) */
	const LADSPA_Data fc = *(plugin_data->fc);

	/* Bandwidth (octaves) (float value) */
	const LADSPA_Data bw = *(plugin_data->bw);

	/* Input (array of floats of length sample_count) */
	const LADSPA_Data * const input = plugin_data->input;

	/* Output (array of floats of length sample_count) */
	LADSPA_Data * const output = plugin_data->output;

	biquad * filter = plugin_data->filter;
	float fs = plugin_data->fs;

	unsigned long pos;

// rt 5.2.2013: set FTZ (flush-to-zero) and DAZ (denormals-are-zero) bits
// in the floating point processor, to avoid denormals slowing us down
// http://www.mattwalsh.com/twiki/bin/view/Main/DenormalNumberTest
// http://code.google.com/p/amsynth/source/browse/src/Effects/denormals.h
  int oldMXCSR = _mm_getcsr(); //read the old MXCSR setting
  int newMXCSR = oldMXCSR | 0x8040; // set DAZ and FZ bits
  _mm_setcsr( newMXCSR ); //write the new MXCSR setting to the MXCSR

	ap_set_params(filter, fc, bw, fs);

	for (pos = 0; pos < sample_count; pos++) {
	  buffer_write(output[pos], biquad_run(filter, input[pos]));
	}

  //restore old MXCSR settings to turn denormals back on if they were on
  _mm_setcsr( oldMXCSR );

}

void _init() {
	char **port_names;
	LADSPA_PortDescriptor *port_descriptors;
	LADSPA_PortRangeHint *port_range_hints;

#define D_(s) (s)

	allPassDescriptor =
	 (LADSPA_Descriptor *)malloc(sizeof(LADSPA_Descriptor));

	if (allPassDescriptor) {
		allPassDescriptor->UniqueID = 9005;
		allPassDescriptor->Label = "RTallpass2";
		allPassDescriptor->Properties =
		 LADSPA_PROPERTY_HARD_RT_CAPABLE;
		allPassDescriptor->Name =
		 D_("RT Biquad Allpass");
		allPassDescriptor->Maker =
		 "Richard Taylor <rtaylor@tru.ca>";
		allPassDescriptor->Copyright =
		 "GPL";
		allPassDescriptor->PortCount = 4;

		port_descriptors = (LADSPA_PortDescriptor *)calloc(4,
		 sizeof(LADSPA_PortDescriptor));
		allPassDescriptor->PortDescriptors =
		 (const LADSPA_PortDescriptor *)port_descriptors;

		port_range_hints = (LADSPA_PortRangeHint *)calloc(4,
		 sizeof(LADSPA_PortRangeHint));
		allPassDescriptor->PortRangeHints =
		 (const LADSPA_PortRangeHint *)port_range_hints;

		port_names = (char **)calloc(4, sizeof(char*));
		allPassDescriptor->PortNames =
		 (const char **)port_names;

		/* Parameters for Frequency (Hz) */
		port_descriptors[ALLPASS_FC] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[ALLPASS_FC] =
		 D_("Frequency (Hz)");
		port_range_hints[ALLPASS_FC].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_SAMPLE_RATE | LADSPA_HINT_DEFAULT_440;
		port_range_hints[ALLPASS_FC].LowerBound = 0;
		port_range_hints[ALLPASS_FC].UpperBound = 0.4;

		/* Parameters for Bandwidth (octaves) */
		port_descriptors[ALLPASS_BW] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_CONTROL;
		port_names[ALLPASS_BW] =
		 D_("Bandwidth (octaves)");
		port_range_hints[ALLPASS_BW].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE | LADSPA_HINT_DEFAULT_1;
		port_range_hints[ALLPASS_BW].LowerBound = 0;
		port_range_hints[ALLPASS_BW].UpperBound = 4;

		/* Parameters for Input */
		port_descriptors[ALLPASS_INPUT] =
		 LADSPA_PORT_INPUT | LADSPA_PORT_AUDIO;
		port_names[ALLPASS_INPUT] =
		 D_("Input");
		port_range_hints[ALLPASS_INPUT].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[ALLPASS_INPUT].LowerBound = -1.0;
		port_range_hints[ALLPASS_INPUT].UpperBound = +1.0;

		/* Parameters for Output */
		port_descriptors[ALLPASS_OUTPUT] =
		 LADSPA_PORT_OUTPUT | LADSPA_PORT_AUDIO;
		port_names[ALLPASS_OUTPUT] =
		 D_("Output");
		port_range_hints[ALLPASS_OUTPUT].HintDescriptor =
		 LADSPA_HINT_BOUNDED_BELOW | LADSPA_HINT_BOUNDED_ABOVE;
		port_range_hints[ALLPASS_OUTPUT].LowerBound = -1.0;
		port_range_hints[ALLPASS_OUTPUT].UpperBound = +1.0;

		allPassDescriptor->activate = activateAllPass;
		allPassDescriptor->cleanup = cleanupAllPass;
		allPassDescriptor->connect_port = connectPortAllPass;
		allPassDescriptor->deactivate = NULL;
		allPassDescriptor->instantiate = instantiateAllPass;
		allPassDescriptor->run = runAllPass;
		allPassDescriptor->run_adding = NULL;
	}
}

void _fini() {
	if (allPassDescriptor) {
		free((LADSPA_PortDescriptor *)allPassDescriptor->PortDescriptors);
		free((char **)allPassDescriptor->PortNames);
		free((LADSPA_PortRangeHint *)allPassDescriptor->PortRangeHints);
		free(allPassDescriptor);
	}

}

